"""
Implementation for a tree data structure to save
CATH information
"""
class CATHTree:
    """
    Generic tree data structure for CATH
    
    It is initialized by one of the ID in the C.A.T.H. code.
    If `cath_code = 0` then the node is considered as the
    root.
    """
    def __init__(self, cath_code: str) -> None:
        if cath_code == "0":
            self.root = True
            self.cath_code = ""
        else:
            self.root = False
            self.cath_code = cath_code
        # to store other CATHTree instances
        self.branches = []
        # to store CATHDomain instances
        self.domains = []

    def get_cath(self, as_list=False) -> str:
        """
        CATH code following branches up until the root.
        One can return the code also as a list.
        """
        if as_list:
            return self.cath_code.split(".")
        return self.cath_code

    def set_cath(self, code: str) -> None:
        """Set a new value for the CATH node (C.A.T.H.)"""
        self.cath_code = code

    ######## Methods for branches ########

    def add_branch(self, node):
        """
        Add a new CATHTree as a branch

        One adds to the current branch (e.g. with code C.)
        the next part of the CATH code (in the example, A.)
        """
        assert isinstance(node, CATHTree)
        # Update cath code
        if not node.is_single_code():
            raise ValueError(f"You have to provide a node with a compound CATH code. Are you sure that you want to add this? Provided CATH code: {node.get_cath()}")
        if not self.root:
            node.set_cath(f"{self.get_cath()}.{node.get_cath()}")
        self.branches.append(node)

    def get_branches(self, only_last_cath_code=False) -> list:
        """List of the branches in the current node"""
        if only_last_cath_code:
            return [b.get_cath(as_list=True)[-1] for b in self.branches]
        return self.branches

    def get_branch_index_from_code(self, last_cath_code) -> int:
        """Branch index search by the last inserted ID of the CATH"""
        try:
            idx = self.get_branches(only_last_cath_code=True).index(last_cath_code)
        except ValueError:
            idx = None
        return idx

    def is_single_code(self) -> bool:
        """Internal function"""
        return len(self.get_cath(as_list=True)) == 1

    def is_cath_code_complete(self) -> bool:
        """Check if all C.A.T.H. IDs are present"""
        return len(self.get_cath(as_list=True)) == 4

    ######## Methods for Domains ########

    def add_domain(self, domain):
        """Add a domain to the node"""
        assert isinstance(domain, CATHDomain)
        self.domains.append(domain)

    def del_domain(self, idx2remove):
        """Remove domain from node by index"""
        del self.domains[idx2remove]

    def get_domains(self) -> list:
        """List of Domains of the node"""
        return self.domains

class CATHDomain:
    """Info regarding CATH domains from domain-list"""
    def __init__(self, domain_name, cath_classification, resolution) -> None:
        assert isinstance(cath_classification, list)
        assert len(cath_classification) == 5
        assert resolution != '', print(domain_name)
        self.domain_info = domain_name
        self.cath_code = cath_classification[:4]
        self.homology35 = cath_classification[4]
        self.resolution = resolution

    def get_homology35_id(self) -> str:
        return self.homology35

    def get_resolution(self) -> str:
        return self.resolution

    def get_domain_info(self) -> str:
        return self.domain_info

    def get_domain_cath(self) -> list:
        return self.cath_code
