"""Dataset classification.

In the following, analysis of the dataset is performed
regarding the following aspects:
- resolution
- trans/mono classification
- intrinsic/peripheral classification
- species
"""
import csv
from collections import Counter
import json
import pathlib

import numpy as np
import matplotlib
import matplotlib.pyplot as plt

from utils import remove_right_upper_axes, fancy_grid, refined_text, set_ticks, check_for_word

parent = pathlib.Path(__file__).parent.resolve()

# load data
with open(parent / "../2_35_homology_filter/polymer_membrane_entities_cath_35homology_filtered.csv", "r", encoding="utf-8") as data_file:
    reader = csv.reader(data_file)
    dataset = list(reader)

# load manually curated list
with open(parent / "manually_curated_entities.txt", 'r', encoding='utf-8') as curatedfile:
    reader = csv.reader(curatedfile)
    curated = {}
    for row in reader:
        if not row[0].startswith("#"):
            curated[row[0]] = row[1:]

# Open file to save list of TRANS/MONO membrane proteins
trans_mono_file = open(
    parent / "classification" /"transmembrane_monotopic_list.csv", "w",
    encoding="utf-8"
)
csvwriter = csv.writer(trans_mono_file)
csvwriter.writerow(["# entity_id", "classification"])

# Open file to save list of INTEGRAL/PERIPHERAL membrane proteins
int_peri_file = open(
    parent / "classification" / "intrinsic_peripheral_list.csv", "w",
    encoding="utf-8"
)
int_peri_writer = csv.writer(int_peri_file)
int_peri_writer.writerow(["# entity_id", "classification"])

# Open file to save list of species
species_file = open(parent / "classification" / "species_list.csv", "w", encoding="utf-8")
species_writer = csv.writer(species_file)
species_writer.writerow(["# entity_id", "classification"])

# Loop over dataset and collect relevant information
resolution = []
number_NMR = 0
# MemProtMD is not considered because it collects integral proteins,
# which can be trans-membrane but not always
number_trans = np.zeros(4) # same order as csv: PDBTM, mpstruc, OPM, manually_curated
number_mono = np.zeros(4)
number_intrinsic = 0
number_peripheral = 0
source_organism = []
other_counter = 0
other_intr_peri_counter = 0
for row in dataset:
    # skip header
    if row[0] == "Entity ID":
        continue

    # resolution
    if row[2]:
        # when multiple resolution are present (Cryo-EM), take the first
        resolution.append(float(row[2].split(",")[0]))
    else:
        # NMR case
        assert "NMR" in row[1], print(row)
        number_NMR += 1

    # membrane classification
    is_curated = False
    # Handle special cases
    if row[0] in curated:
        if curated[row[0]][0] == 'trans':
            number_trans[3] += int(row[15])
            is_trans = True
        elif curated[row[0]][0] == 'mono':
            is_mono = True
            number_mono[3] += int(row[15])
        else:
            raise RuntimeError(f"Error in reading curated file for row: {row}")
        is_curated = True
    # trans/mono
    if not is_curated:
        is_trans = False
        if row[9]:
            # PDBTM field
            # If it is not empty, than its probable a trans-membrane protein.
            # Nevertheless, check for a confirm
            if (row[11] or row[12]) and (check_for_word("transmembrane", row[11]) or check_for_word("transmembrane", row[12])):
                number_trans[1] += int(row[15])
            elif row[13] and check_for_word("transmembrane", row[13]):
                number_trans[2] += int(row[15])
            # check
            elif check_for_word("monotopic", row[11]) or check_for_word("monotopic", row[12]) or check_for_word("monotopic", row[13]):
                raise RuntimeError(f"Row classified by PDBTM by has monotopic key word on other database: {row}\n")
            else:
                # case in which only the PDBTM entry is present
                number_trans[0] += int(row[15])
            is_trans = True
        elif (row[11] or row[12]) and (check_for_word("transmembrane", row[11]) or check_for_word("transmembrane", row[12])):
            # search only inside the other two databases
            number_trans[1] += int(row[15])
            is_trans = True
        elif row[13] and check_for_word("transmembrane", row[13]):
            number_trans[2] += int(row[15])
            is_trans = True

        # Monotopic search
        is_mono = False
        if check_for_word("monotopic", row[11]) or check_for_word("monotopic", row[12]):
            if is_trans:
                raise RuntimeError(f"Entry already classified as trans is mono too: {row}")
            number_mono[1] += int(row[15])
            is_mono = True
        elif check_for_word("monotopic", row[13]):
            if is_trans:
                raise RuntimeError(f"Entry already classified as trans is mono too: {row}")
            number_mono[2] += int(row[15])
            is_mono = True
        ######################### TODO #########################
        # Actually, AMP can be both transmembrane and monotopic, even concurrently!
        # Just depends on the context
        ########################################################
        # Peptides case: some peptides are considered monotopic and
        # interact with the membrane. An example are antimicrobial
        # peptides, used by the innate immune system to treat infections
        # Example: 1ZRX
        # elif check_for_word("peptides", row[11]) or check_for_word("peptides", row[12]):
        #     number_mono[1] += 1
        #     is_mono = True
        # elif check_for_word("peptides", row[13]):
        #     number_mono[2] += 1
        #     is_mono = True
        ########################################################
    # check
    if not is_mono and not is_trans:
        print(f"WARNING TRANSMONO: Row is neither mono nor trans: {row}")
    # save outfile
    if is_trans:
        csvwriter.writerow([f"{row[0][:4]}_{row[7]}", "trans", json.loads(row[17].replace("'", "\""))])
    elif is_mono:
        csvwriter.writerow([f"{row[0][:4]}_{row[7]}", "mono", json.loads(row[17].replace("'", "\""))])
    else:
        csvwriter.writerow([f"{row[0][:4]}_{row[7]}", "other", json.loads(row[17].replace("'", "\""))])
        other_counter += int(row[15])
    # else:
    #     raise RuntimeError(f"The following row has not been classified as trans or mono: {row}")

    # Save Source organism
    source_organism.append(row[8])
    species_writer.writerow([f"{row[0][:4]}_{row[7]}", row[8]])

    # intrinsic peripheral
    is_intrinsic = False
    is_peripheral = False
    is_curated = False
    # Handle special cases
    if row[0] in curated:
        if curated[row[0]][0] == 'trans':
            number_intrinsic += int(row[15])
            is_intrinsic = True
        elif curated[row[0]][0] == 'mono':
            if curated[row[0]][1] == 'intr':
                number_intrinsic += int(row[15])
                is_intrinsic = True
            elif curated[row[0]][1] == 'peri':
                number_peripheral += int(row[15])
                is_peripheral = True
            else:
                raise RuntimeError(f"No classification peri/intr for mono protein in curated: {curated[row[0]]  }")
        else:
            raise RuntimeError(f"Error in reading curated file for row: {row}")
        is_curated = True
    if not is_curated:
        if row[9] or row[10]:
            # PDBTM and MemProtMD field
            # If it is not empty, than its probable a trans-membrane protein.
            # Nevertheless, check for a confirm
            number_intrinsic += int(row[15])
            is_intrinsic = True
        elif (row[11] or row[12]) and (check_for_word("transmembrane", row[11]) or check_for_word("transmembrane", row[12])):
            # search only inside the other two databases
            number_intrinsic += int(row[15])
            is_intrinsic = True
        elif row[13] and check_for_word("transmembrane", row[13]):
            number_intrinsic += int(row[15])
            is_intrinsic = True

        # peripheral search
        if check_for_word("peripheral", row[11]) or check_for_word("peripheral", row[12]):
            if is_intrinsic:
                raise RuntimeError(f"Entry already classified as integral is peripheral too: {row}")
            number_peripheral += int(row[15])
            is_peripheral = True
        elif check_for_word("peripheral", row[13]):
            if is_intrinsic:
                raise RuntimeError(f"Entry already classified as integral is peripheral too: {row}")
            number_peripheral += int(row[15])
            is_peripheral = True
    # check
    if not is_peripheral and not is_intrinsic:
        print(f"WARNING INTPERI: Row is neither Integral nor Peripheral: {row}")
        other_intr_peri_counter += int(row[15])
    # save outfile
    if is_intrinsic:
        int_peri_writer.writerow([f"{row[0][:4]}_{row[7]}", "intr", json.loads(row[17].replace("'", "\""))])
    elif is_peripheral:
        int_peri_writer.writerow([f"{row[0][:4]}_{row[7]}", "peri", json.loads(row[17].replace("'", "\""))])
    else:
        int_peri_writer.writerow([f"{row[0][:4]}_{row[7]}", "other", json.loads(row[17].replace("'", "\""))])

# Closing files
trans_mono_file.close()
int_peri_file.close()
species_file.close()
# ==================================================== #
# ===================== PLOTS ======================== #
# ==================================================== #
matplotlib.rcParams['font.size'] = 24
matplotlib.rcParams['text.usetex'] = True
matplotlib.rcParams['font.family'] = 'modern computer'

# ==================================================== #
# Resolution distribution
fig, ax = plt.subplots(1, figsize=(10,10))

edges = np.linspace(0, 5, 20+1,endpoint=True)
ax.hist(resolution, bins=edges, rwidth=0.9, color="#3498DB")
refined_text(ax, f"$N = {len(resolution)}$\n$N_"+r"{NMR}"+f" = {number_NMR}$", x=0.8, y=0.85)

# refinements
remove_right_upper_axes(ax)
set_ticks(ax, {'x_major': 1, 'x_minor': 0.5, 'y_major': 25})
fancy_grid(ax, {'y_major': True})
ax.set_title("Resolution distribution of membrane chains", pad=15)
fig.savefig(parent / "plots" / "resolution_distribution.pdf", bbox_inches='tight')

# ==================================================== #
# Monotopic/Trans-M
fig, ax = plt.subplots(1, figsize=(10,10))

classes = ["Bitopic (Trans-)", "Monotopic"]
databases = ["PDBTM", "mpstruc", "OPM", "Manually curated"]
heat_map = np.vstack((number_trans, number_mono))
im = ax.imshow(heat_map, cmap="coolwarm")

# Show all ticks and label them with the respective list entries
ax.set_xticks(np.arange(len(databases)), labels=databases)
ax.set_yticks(np.arange(len(classes)), labels=classes)

# Rotate the tick labels and set their alignment.
plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
         rotation_mode="anchor")

# Loop over data dimensions and create text annotations
for i in range(len(classes)):
    for j in range(len(databases)):
        text = ax.text(j, i, f"{heat_map[i, j]:.0f}", ha="center", va="center", color="w")

# refinements
ax.tick_params(
    axis='x',
    which='both',
    bottom=False,
    top=False
)
ax.tick_params(
    axis='y',
    which='both',
    left=False
)      
ax.set_title("Trans/mono domains", pad=15)
fig.savefig(parent / "plots" / "trans_mono.pdf", bbox_inches='tight')

# ==================================================== #
# Integral/peripheral
fig, ax = plt.subplots(1, figsize=(10,10))

classes = ["Integral", "Peripheral"]
databases = ["Number"]
heat_map = np.vstack((number_intrinsic, number_peripheral))
im = ax.imshow(heat_map, cmap="coolwarm")

# Show all ticks and label them with the respective list entries
ax.set_xticks(np.arange(len(databases)), labels=databases)
ax.set_yticks(np.arange(len(classes)), labels=classes)

# Rotate the tick labels and set their alignment.
plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
         rotation_mode="anchor")

# Loop over data dimensions and create text annotations
for i in range(len(classes)):
    for j in range(len(databases)):
        text = ax.text(j, i, f"{heat_map[i, j]:.0f}", ha="center", va="center", color="w")

# refinements
ax.tick_params(
    axis='x',
    which='both',
    bottom=False,
    top=False
)
ax.tick_params(
    axis='y',
    which='both',
    left=False
)      
ax.set_title("Integral/Peripheral domains", pad=15)
fig.savefig(parent / "plots" / "integral_peripheral.pdf", bbox_inches='tight')

# ==================================================== #
# source organism
fig, ax = plt.subplots(1, figsize=(20,10))

# counting
c = Counter(source_organism)
# sorting
dict_sorted = dict(sorted(c.items(), key=lambda item: item[1]))
# plotting
percentage2plot = 0.1
end_2plot = int(0.9*len(dict_sorted))
ax.bar(list(dict_sorted.keys())[:end_2plot:-1] , list(dict_sorted.values())[:end_2plot:-1])

# refinements
remove_right_upper_axes(ax)
plt.setp(ax.get_xticklabels(), rotation=90,
         ha="right", va='center',
         rotation_mode="anchor")
fancy_grid(ax, {'y_major': True})
ax.set_ylabel('Counts')
ax.set_title("Source Organism", pad=15)
fig.savefig(parent / "plots" / "source_organism.pdf", bbox_inches='tight')

print(f"Number of others: {other_counter}")
print(f"Number of others (intr/peri): {other_intr_peri_counter}")
